/*
    Copyright (C) 2000 by Anders Larsen and Baumer Ident GmbH

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program. If not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <asm/types.h>
#include <sys/stat.h>

int verbose = 0;

__u8  signature[7] = "B000FF\n";
__u32 start        =    0x90000;
__u32 test         =   0x17db90;


int usage( char* const argv0 )
{
  fprintf( stderr, "use: %s [ -<option> ]* initrd\n", argv0 );
  fputs( "\noptions:\n",                                     stderr );
  fputs( "\tm <size> : memory size(dec) in kb\n",            stderr );
  fputs( "\t           default: from 'boot'\n",              stderr );
  fputs( "\tr <addr> : load addr(hex) of initrd\n",          stderr );
  fputs( "\t           default: highest possible address\n", stderr );
  fputs( "\tv        : be verbose\n",                        stderr );
  return EXIT_FAILURE;
}


int copy( char const* name, __u32 offset, __u32 addr, FILE* fpo )
{
  int   ch;
  __u32 cksum = 0;
  __u32 size  = 0;
  FILE* fpi;

  if ( !( fpi = fopen( name, "r" ) ) )
  {
    perror( "can't open input file" );
    return EXIT_FAILURE;
  }
  fseek( fpo, 3 * sizeof( __u32 ), SEEK_CUR );
  if ( offset )
    fseek( fpi, offset, SEEK_SET );
  while ( ( ch = getc( fpi ) ) != EOF )
  {
    size++;
    cksum += ch;
    putc( ch, fpo );
  }
  fseek ( fpo, -(long)( size + 3 * sizeof( __u32 ) ), SEEK_CUR );
  fwrite( &addr,  sizeof( addr  ), 1, fpo );
  fwrite( &size,  sizeof( size  ), 1, fpo );
  fwrite( &cksum, sizeof( cksum ), 1, fpo );
  fseek ( fpo, size, SEEK_CUR );

  if ( verbose )
    printf( "%16s @ %08x, %7d bytes (checksum %08x)\n", name, addr, size, cksum );

  fclose( fpi );
  return size;
}


int main( int argc, char* const* argv )
{
  int   ch;
  __u16 offset;
  __u32 memsize  = 0;
  __u32 rootaddr = 0;
  __u32 rootsize;
  __u32 linuxoff;
  __u32 ul;
  char  version[41];
  FILE* fpi;
  FILE* fpo;
  char const* cp;
  struct stat statbuf;

  while ( ( ch = getopt( argc, argv, "vm:r:" ) ) != EOF )
    switch ( ch )
    {
      case 'm':
        memsize  = strtol( optarg, 0, 10 ) - 1024;
        break;

      case 'r':
        rootaddr = strtol( optarg, 0, 16 );
        break;

      case 'v':
        verbose++;
        break;

      default:
        return usage( argv[0] );
    }

  if ( optind != argc - 1 )
    return usage( argv[0] );

  // get size of initrd
  if ( stat( argv[optind], &statbuf ) )
  {
    perror( "can't stat initrd" );
    return EXIT_FAILURE;
  }
  rootsize = statbuf.st_size;

  // extract information from "bzImage"
  if ( !( fpi = fopen( "bzImage", "r" ) ) )
  {
    perror( "can't open bzImage" );
    return EXIT_FAILURE;
  }
  fseek ( fpi, 0x1f1, SEEK_SET );
  fread ( &linuxoff, sizeof( linuxoff ), 1, fpi );
  fseek ( fpi, 0x20e, SEEK_SET );
  fread ( &offset,   sizeof( offset ),   1, fpi );
  fseek ( fpi, offset + 0x200, SEEK_SET );	// SETUPSEG - INITSEG
  fread ( version,   sizeof( version ),  1, fpi );
  version[sizeof( version ) - 1] = 0;
  fclose( fpi );
  linuxoff = ( ( linuxoff & 0xff ) + 1 ) * 512;

  // update information in "boot"
  if ( !( fpi = fopen( "boot", "r+" ) ) )
  {
    perror( "can't open boot" );
    return EXIT_FAILURE;
  }
  // copy kernel version string
  if ( ( cp = strtok( version, " " ) ) != NULL )
  {
    fseek ( fpi, 0x22e, SEEK_SET );			// 0x20e + Minix header
    fread ( &offset,   sizeof( offset ), 1, fpi );
    fseek ( fpi, offset + 0x26, SEEK_SET );	// Minix header + "Linux "
    fwrite( cp, strlen( cp ) + 1, 1, fpi );  
  }
  // get memory size
  fseek ( fpi, 0x246, SEEK_SET );			// 0x226 + Minix header
  if ( memsize == 0 )
    fread ( &memsize, sizeof( memsize ), 1, fpi );
  else
    fwrite( &memsize, sizeof( memsize ), 1, fpi );
  memsize = ( memsize + 1024 ) * 1024;
  // write initrd info
  if ( rootaddr == 0 )
    rootaddr = ( memsize - rootsize ) & 0xfffe0000;
  fseek ( fpi, 0x238, SEEK_SET );
  fwrite( &rootaddr, sizeof( rootaddr ), 1, fpi );
  fwrite( &rootsize, sizeof( rootsize ), 1, fpi );
  fclose( fpi );

  // write NK.bin header
  if ( !( fpo = fopen( "nk.bin", "w+" ) ) )
  {
    perror( "can't open output file" );
    return EXIT_FAILURE;
  }
  fwrite( signature, sizeof( signature ), 1, fpo );
  fwrite( &start,    sizeof( test      ), 1, fpo );
  fwrite( &test,     sizeof( test      ), 1, fpo );
  
  copy( "boot",          32,  0x90000, fpo );
  copy( "bzImage", linuxoff, 0x100000, fpo );
  copy( argv[optind],     0, rootaddr, fpo );

  // write NK.bin trailer
  ul = 0;
  fwrite( &ul, sizeof( ul ), 1, fpo );
  ul = 0x90000;
  fwrite( &ul, sizeof( ul ), 1, fpo );
  ul = 0;
  fwrite( &ul, sizeof( ul ), 1, fpo );

  fclose( fpo );
  return EXIT_SUCCESS;
}
